#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include "lzpipe.h"
#include "xla2.h"
#include "rle.h"

/* Type that describes an XLA's header */
typedef struct THeader {
  char sig[4];							/* Signature, always "XLAS" */
  long posdir, sizedir;			/* Position and length of XLA directory */
} THeader;

typedef struct TFile {
  char name[64];						/* Filename */
  long posfile;							/* Position of the file */
  long sizefile;						/* Size of the original file */
  long sizecomp;						/* Compressed size */
  long algorithm;						/* Algorithm used : None, LZS, Other */
} TFile;

typedef struct TXLADir {		/* XLA's directory memory format */
  TFile item;
  struct TXLADir *next;
} TXLADir;


static FILE *XLAFile;				/* Handle of the main file */
static THeader Header;			/* Header for the current file */
static TXLADir *XLADir;			/* XLA directory */
static TXLADir *CurrentDir; /* */
static long BytesWritten;		/* Amount of bytes written to disk */
static char SearchPattern[255];	/* Pattern for search matching */

#define BUFSIZE 4096				/* Size of disk buffer */
static char InBuf[BUFSIZE]; /* Input buffer */
static char OutBuf[BUFSIZE];/* Output buffer */
static int XWriteSize;
short ratio;
unsigned short ModeUsed;
int (*GetChar)();
void (*PutChar)(int);    

static boolean ArchiveOpen; /* Flag that defines if an archive is open or not */

void FSplit( s, n, e )
char *s;
char *n;
char *e;
{
	char *tmp1;
	char *tmp2;
	char *tmp3;
	
	tmp1 = strchr(s,'.');
	for(tmp2=s,tmp3=n;tmp2<tmp1;tmp2++,tmp3++)
		*tmp3 = *tmp2;
	*(tmp3) = '\0';
	strcpy(e,(void *) (tmp1+1));
}

int strpos( s, c )
char *s;
char c;
{
	int i;
	i = 0;
	while (s[i]!='\0') {
		if (s[i]==c) return(i);
		i++;
	}
	return(-1);
}

/* Function that cleans up various memory structures that have been allocated */
static void CleanUpAll()
{
  TXLADir *tmp;

  while (XLADir != NULL) {
    tmp = XLADir->next;
    free(XLADir);
    XLADir = tmp;
  }
}

/* Recursive function that adds an entry to an archive */
static void AddName( TXLADir **P, TXLADir **Q)
{
  if (*P != NULL)
    AddName(&(*P)->next, Q);
  else
    *P = *Q;
}

/* Function that creates an archive */
boolean XCreateArchive(char *filename)
{
	/* Try to create an archive */
  if (!(XLAFile = fopen(filename, "wb")))
    return FALSE;
  
  /* Set up the header's structure */
  memcpy(Header.sig, "XLAS", 4L);
  Header.posdir = sizeof(THeader);
  Header.sizedir = 0;
  
  /* Write the header to disk */
  fwrite(&Header, sizeof(THeader), 1, XLAFile);
  XLADir = NULL;
  ArchiveOpen = TRUE;
  return TRUE;
}

/* Function that closes an archive */
boolean XEndArchive()
{
  TXLADir *tmp;
	
	/* Check if an archive is currently open */
  if (!ArchiveOpen)
    return FALSE;
    
  /* Go to the beginning of the XLA's directory */
  fseek(XLAFile, Header.posdir, 0);
  
  /* Traverse the XLA's directory, writing it to disk */
  tmp = XLADir;
  while (tmp != NULL) {
    fwrite(&tmp->item, sizeof(TFile), 1, XLAFile);
    tmp = tmp->next;
  }
  
  /* Go to beginning of file */
  fseek(XLAFile, 0L, 0);
  
  /* Write the header and close the file */
  fwrite(&Header, sizeof(THeader), 1, XLAFile);
  fclose(XLAFile);
  
  /* Clean up status variables and return successfully */
  CleanUpAll();
  ArchiveOpen = FALSE;
  return TRUE;
}

void XWriteChar(int c)
{
	putc(c,XLAFile);
	XWriteSize++;
}

int XReadChar()
{
	return(getc(XLAFile));
}

/* Function that adds a new file to an archive */
boolean XLAPut(char *fname, unsigned int mode, int (*XGetChar)() )
{
  TXLADir *tmp;
  short i;
  int c, buff, NotFinished, TotalSize;
  

  /* Check if an archive is open */
  if (!ArchiveOpen)
    return FALSE;
    
  /* Increment the number of files and allocate a new directory entry */
  Header.sizedir++;
  tmp = (TXLADir *)malloc(sizeof(TXLADir));
  tmp->next = NULL;
  tmp->item.posfile = Header.posdir;
  
  /* Pad a string out with spaces */
  /*for (i = 0; i <= 11; i++) {
    if (i + 1 <= strlen(fname))
      tmp->item.name[i] = fname[i];
    else
      tmp->item.name[i] = ' ';
  }*/
  strcpy(tmp->item.name,fname);

  BytesWritten = 0;
  TotalSize = 0;

  fseek(XLAFile, Header.posdir, 0);
  switch (mode) {

  case None:
    while ((c=(XGetChar)())!=EOF) {
      putc(c, XLAFile);
      BytesWritten++;
    }
    TotalSize = BytesWritten;
    ModeUsed = None;
    break;
  case RLE:
  	BytesWritten = WriteRLE(XLAFile,XGetChar,&TotalSize);
  	ModeUsed = RLE;
  	break;
  case DeflateFast:
  	XWriteSize = 0;
  	buff = 0;
  	NotFinished = 1;
  	zipcreat(XWriteChar, ZIP_GNU, 1);
  	while (1) {
  		while (((c=(XGetChar)())!=EOF)&&(buff<2048))
  			InBuf[buff++] = (unsigned char) c;
  		zipwrite(InBuf,buff);
  		TotalSize += buff;
  		if (c==EOF) break;
  		buff = 0;
  	} 
  	zipclose();
  	BytesWritten = XWriteSize;
  	ModeUsed = DeflateFast;
  	break;
  	
  case DeflateLazy:
  	XWriteSize = 0;
  	buff = 0;
  	NotFinished = 1;
  	zipcreat(XWriteChar, ZIP_GNU, 9);
  	while (1) {
  		memset(InBuf,0,BUFSIZE);
  		while (((c=(XGetChar)())!=EOF)&&(buff<2048))
  			InBuf[buff++] = (unsigned char) c;
  		zipwrite(InBuf,buff);
  		TotalSize += buff;
  		if (c==EOF) break;
  		buff = 0;
  	} 
  	zipclose();
  	BytesWritten = XWriteSize;
  	ModeUsed = DeflateLazy;
  	break;
  		
  	
	default:
		break;
  }
  tmp->item.sizefile = TotalSize;
  tmp->item.sizecomp = BytesWritten;
  tmp->item.algorithm = ModeUsed;
  ratio = 100 - BytesWritten * 100 / TotalSize;
  Header.posdir += BytesWritten;
  tmp->next = NULL;
  AddName(&XLADir, &tmp);
  return TRUE;
}


boolean XLAGet(char *fname, void (*XPutChar)(int))
{
  short i;
  char name[64];
  TXLADir *tmp;
  long TotalSize;
  int InBufSize;
  int c;

  if (!ArchiveOpen)
    return FALSE;
  /*
  for (i = 0; i <= 11; i++) {
    if (i + 1 <= strlen(fname))
      name[i] = fname[i];
    else
      name[i] = ' ';
  }*/
  strcpy(name,fname);

  tmp = XLADir;
  while (strcmp(name, tmp->item.name)) {
    if (tmp == NULL)
      return FALSE;
    tmp = tmp->next;
  }
  fseek(XLAFile, tmp->item.posfile, 0);
  TotalSize = tmp->item.sizecomp;
  switch (tmp->item.algorithm) {

  	case None:
  		InBufSize = 0;
      	while ((InBufSize++)<TotalSize) (XPutChar)(fgetc(XLAFile));
    	ModeUsed = None;
    	break;
    
    case RLE:
    	ReadRLE(XLAFile,TotalSize,XPutChar);
    	ModeUsed = RLE;
    	break;
    	
    case DeflateLazy: case DeflateFast:
    	unzopen(XReadChar, ZIP_GNU);
    	while ((InBufSize=unzread(InBuf,2048))>0) {
    		for(i=0;i<InBufSize;i++) (XPutChar)(InBuf[i]);
    	}	
    	zipclose();
    	ModeUsed = tmp->item.algorithm;
    	break;
    		
		default:
			break;
	}
  return TRUE;
}


boolean XOpenArchive(char *filename)
{
  short i;
  TXLADir *tmp;
  char sig[5];

  if (ArchiveOpen)
    return FALSE;
  if (!(XLAFile = fopen(filename, "r+b"))) 
  	return FALSE;
  fread(&Header, sizeof(THeader), 1, XLAFile);
  if (memcmp(Header.sig, "XLAS", 4))
    return FALSE;
  XLADir = NULL;
  fseek(XLAFile, Header.posdir, 0);
  for (i = 1; i <= Header.sizedir; i++) {
    tmp = (TXLADir *)malloc(sizeof(TXLADir));
    fread(&tmp->item, sizeof(TFile), 1, XLAFile);
    tmp->next = NULL;
    AddName(&XLADir, &tmp);
  }
  ArchiveOpen = TRUE;
  return TRUE;
}


boolean XUpdateArchive(filename)
char *filename;
{
  short i;
  TXLADir *tmp;
  char sig[5];

  if (ArchiveOpen)
    return FALSE;
  if (!(XLAFile = fopen(filename, "r+b"))) 
  	return FALSE;
  fread(&Header, sizeof(THeader), 1, XLAFile);
  strcpy(sig, "XLAS");
  if (strncmp(Header.sig, sig, 4))
    return FALSE;
  XLADir = NULL;
  fseek(XLAFile, Header.posdir, 0);
  for (i = 1; i <= Header.sizedir; i++) {
    tmp = (TXLADir *)malloc(sizeof(TXLADir));
    fread(&tmp->item, sizeof(TFile), 1, XLAFile);
    tmp->next = NULL;
    AddName(&XLADir, &tmp);
  }
  fseek(XLAFile, Header.posdir, 0);
  truncate(XLAFile);
  ArchiveOpen = TRUE;
  return TRUE;
}


boolean XCloseArchive()
{
  if (!ArchiveOpen)
    return FALSE;
  else {
    if (XLAFile != NULL)
      fclose(XLAFile);
    XLAFile = NULL;
    CleanUpAll();
    ArchiveOpen = FALSE;
    return TRUE;
  }
}


void XPrintDir()
{
  TXLADir *tmp;
  char s[256];
  long totsize, totcomp;
  TFile *WITH;

  if (!ArchiveOpen)
    return;
  printf("Name                Size       CSize   Ratio    Position   Method\n");
  printf("-----------------------------------------------------------------\n");
  tmp = XLADir;
  totsize = 0;
  totcomp = 0;
  while (tmp != NULL) {
    s[12] = '\0';
    memmove(s, tmp->item.name, 12L);
    WITH = &tmp->item;
    printf("%12s%12ld%12ld%8.2f%12ld",
	   s, WITH->sizefile, WITH->sizecomp,
	   100 - WITH->sizecomp * 100.0 / WITH->sizefile, WITH->posfile);
    switch (WITH->algorithm) {

    case None:
      printf("   Stored\n");
      break;

    case LZS:
      printf("      LZS\n");
      break;
    
    case DeflateLazy:
    	printf(" Deflate9\n");
    	break;
    
    case DeflateFast:
    	printf(" Deflate1\n");
    	break;
    
    case RLE:
    	printf("      RLE\n");
    	break;

    default:
      printf("  Unknown\n");
      break;
    }
    totsize += WITH->sizefile;
    totcomp += WITH->sizecomp;
    tmp = tmp->next;
  }
  *s = '\0';
  printf("-----------------------------------------------------------------\n");
  printf("%12s%12ld%12ld%8.2f\n",
	 s, totsize, totcomp, 100 - totcomp * 100.0 / totsize);
}


boolean XLAFindNext(match)
char *match;
{
  char n1[9], n2[9];
  char e1[4], e2[4];
  char filename[80];
  short i;
  int wildname, wildext;
  char prefixname[80], prefixext[80];
  boolean matchname, matchext;
  char STR1[256];

  FSplit(SearchPattern, n1, e1);
  wildname = strpos(n1, '*')+1;
  wildext = strpos(e1, '*')+1;
  sprintf(prefixname, "%.*s", wildname - 1, n1);
  sprintf(prefixext, "%.*s", wildext - 1, e1);

  while (CurrentDir != NULL) {
    memmove(filename, CurrentDir->item.name, 12L);
    i = 0;
    while (i <= 11 && CurrentDir->item.name[i] != ' ')
      i++;
    filename[i] = '\0';
    FSplit(filename, n2, e2);
    if (*e2 == '\0')
      strcpy(e2, ".");
    matchname = (wildname == 0 && !strcmp(n1, n2) ||
		 (wildname > 0 &&
		  !strcmp((sprintf(STR1, "%.*s", wildname - 1, n2), STR1),
			  prefixname)));
    matchext = (wildext == 0 && !strcmp(e1, e2) ||
			(wildext > 0 &&
	 		!strcmp((sprintf(STR1, "%.*s", wildext - 1, e2), STR1), prefixext)));
    if (matchname && matchext) {
      strcpy(match, filename);
      CurrentDir = CurrentDir->next;
      return TRUE;
    }
    CurrentDir = CurrentDir->next;
  }
  return FALSE;
}

boolean XLAFindFirst(pattern, match)
char *pattern;
char *match;
{
  CurrentDir = XLADir;
  strcpy(SearchPattern, pattern);
  return (XLAFindNext(match));
}

boolean XLANext(char *match)
{
	while (CurrentDir!=NULL) {
		strcpy(match,CurrentDir->item.name);
		CurrentDir = CurrentDir->next;
		return(TRUE);
	}
	return(FALSE);
}
		
	
boolean XLAFirst(char *match)
{
	CurrentDir = XLADir;
	return(XLANext(match));
}

boolean XLAGetFileInfo( char *fname, long *origsize, long *compsize, int *mode )
{
	TXLADir *tmp;
	int i;
	
	if (!ArchiveOpen)
		return(FALSE);
	tmp = XLADir;
	if (tmp==NULL) 
		return(FALSE);
	while (strncmp(fname,tmp->item.name,strlen(fname))) {
		if (tmp->next==NULL) 
			return(FALSE);
		tmp = tmp->next;
	}
	*origsize = tmp->item.sizefile;
	*compsize = tmp->item.sizecomp;
	*mode = tmp->item.algorithm;
	return(TRUE);
}



void XLA2Init()
{
  XLAFile = NULL;
  ArchiveOpen = FALSE;
  XLADir = NULL;
}


/* End. */
